@@ -1,10 +1,7 @@
 /* cert-gpgsm.c - Scute certificate searching.
-   Copyright (C) 2006 g10 Code GmbH
+   Copyright (C) 2006, 2007 g10 Code GmbH
 
-   This file is part of Scute[1].
-
-   [1] Derived from the RSA Security Inc. PKCS #11 Cryptographic Token
-   Interface (Cryptoki).
+   This file is part of Scute.
  
    Scute is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
@@ -34,7 +31,6 @@
 #include <config.h>
 #endif
 
-#include <errno.h>
 #include <time.h>
 #include <string.h>
 #include <stdlib.h>
@@ -80,6 +76,8 @@ cert_reset (struct cert *cert)
     free (cert->issuer_name);
   if (cert->uid)
     free (cert->uid);
+  if (cert->cert_der)
+    free (cert->cert_der);
 
   memset (cert, '\0', sizeof (struct cert));
 }
@@ -162,7 +160,7 @@ decode_c_string (const char *src, char **destp, size_t len)
 	 string.  */
       dest = malloc (strlen (src) + 1);
       if (!dest)
-	return gpg_error_from_errno (errno);
+	return gpg_error_from_syserror ();
 
       *destp = dest;
     }
@@ -264,9 +262,9 @@ search_certs_line (struct search_ctx *ctx)
 
   /* Strip a trailing carriage return.  */
   if (ctx->pending_len > 0
-      && ctx->pending[ctx->pending_len] == '\r')
+      && ctx->pending[ctx->pending_len - 1] == '\r')
     ctx->pending_len--;
-  ctx->pending[ctx->pending_len] = '\0';
+  ctx->pending[ctx->pending_len - 1] = '\0';
   ctx->pending_len = 0;
 
   cert = &ctx->cert;
@@ -351,7 +349,7 @@ search_certs_line (struct search_ctx *ctx)
 	{
 	  cert->issuer_serial = strdup (field[7]);
 	  if (!cert->issuer_serial)
-	    return gpg_error_from_errno (errno);
+	    return gpg_error_from_syserror ();
 	}
 
 #if 0
@@ -377,8 +375,10 @@ search_certs_line (struct search_ctx *ctx)
     case RT_UID:
       if (cert->valid)
 	{
-	  /* Field 2 has the trust info, and field 10 has the user ID.  */
-	  if (fields >= 10)
+	  /* Field 2 has the trust info, and field 10 has the user ID.
+	     Note that more than one UID field can appear.  We only
+	     remember the last one.  It's not used anyway.  */
+	  if (fields >= 10 && !cert->uid)
 	    {
 	      if (decode_c_string (field[9], &cert->uid, 0))
 		return gpg_error (GPG_ERR_ENOMEM);	/* FIXME */
@@ -519,8 +519,10 @@ struct search_ctx_by_field
 };
   
 
+/* This is a compatibility function for GPGSM 2.0.0, which does not
+   support the --data option with the EXPORT command.  */
 static gpg_error_t
-export_cert (char *fpr, struct cert *cert)
+export_cert_compat (char *fpr, struct cert *cert)
 {
   gpg_error_t err;
   assuan_context_t ctx;
@@ -531,8 +533,13 @@ export_cert (char *fpr, struct cert *cert)
   int output_fds[2];
   int child_fds[2];
 
+#define MAX_CERT_SIZE 4096
+  cert->cert_der = malloc (MAX_CERT_SIZE);
+  if (!cert->cert_der)
+    return gpg_error_from_syserror ();
+
   if(pipe (output_fds) < 0)
-    return gpg_error_from_errno (errno);
+    return gpg_error_from_syserror ();
 
   child_fds[0] = output_fds[1];
   child_fds[1] = -1;
@@ -576,6 +583,105 @@ export_cert (char *fpr, struct cert *cert)
 }
 
 
+struct export_hook
+{
+  /* The exported data.  */
+  char *buffer;
+
+  /* The length of the exported data buffer.  */
+  unsigned int buffer_len;
+
+  /* The size of the allocated exported data buffer.  */
+  unsigned int buffer_size;
+};
+
+#define EXP_DATA_START 4096
+
+static gpg_error_t
+export_cert_cb (void *hook, char *line, size_t line_len)
+{
+  struct export_hook *exp = hook;
+
+  if (exp->buffer_size - exp->buffer_len < line_len)
+    {
+      unsigned int new_buffer_size = exp->buffer_size ?
+	(exp->buffer_size * 2) : EXP_DATA_START;
+      char *new_buffer = realloc (exp->buffer, new_buffer_size);
+
+      if (!new_buffer)
+	return gpg_error_from_syserror ();
+
+      exp->buffer = new_buffer;
+      exp->buffer_size = new_buffer_size;
+    }
+
+  memcpy (exp->buffer + exp->buffer_len, line, line_len);
+  exp->buffer_len += line_len;
+
+  return 0;
+}
+
+
+static gpg_error_t
+export_cert (char *fpr, struct cert *cert)
+{
+  gpg_error_t err;
+  assuan_context_t ctx;
+  const char *argv[] = { "gpgsm", "--server", NULL };
+#define COMMANDLINELEN 80
+  char cmd[COMMANDLINELEN];
+  int output_fds[2];
+  int child_fds[2];
+  struct export_hook exp;
+
+  if(pipe (output_fds) < 0)
+    return gpg_error_from_syserror ();
+
+  child_fds[0] = output_fds[1];
+  child_fds[1] = -1;
+
+  err = assuan_pipe_connect (&ctx, GPGSM_PATH, argv, child_fds);
+  close (output_fds[1]);
+  if (err)
+    {
+      close (output_fds[0]);
+      return err;
+    }
+
+  exp.buffer = NULL;
+  exp.buffer_len = 0;
+  exp.buffer_size = 0;
+
+  snprintf (cmd, sizeof (cmd), "EXPORT --data -- %s\n", cert->fpr);
+  err = assuan_transact (ctx, cmd, export_cert_cb, &exp,
+			 NULL, NULL, NULL, NULL);
+  assuan_disconnect (ctx);
+  close (output_fds[0]);
+
+  if (!err)
+    {
+      cert->cert_der = exp.buffer;
+      cert->cert_der_len = exp.buffer_len;
+    }
+  else if (gpg_err_code (err) == GPG_ERR_ASS_NO_OUTPUT)
+    {
+      /* For compatibility with GPGSM 2.0.0, we fall back to a work
+	 around in that case.  */
+      if (cert->cert_der)
+	{
+	  free (cert->cert_der);
+	  cert->cert_der = NULL;
+	}
+      err = export_cert_compat (fpr, cert);
+    }
+
+  if (!err)
+    err = scute_agent_is_trusted (fpr, &cert->is_trusted);
+
+  return err;
+}
+
+
 static gpg_error_t
 search_certs_by_field (void *hook, struct cert *cert)
 {
@@ -593,8 +699,6 @@ search_certs_by_field (void *hook, struct cert *cert)
 	return err;
 
       err = (*ctx->search_cb) (ctx->search_cb_hook, cert);
-
-      /* If necessary, release cert->der here.  */
     }
 
   return err;
